;--------------------------------------------------------------------
; CENTIPED.ASM
; version 1.0
; (C) december 2006
;--------------------------------------------------------------------

ROM_GET		equ	$c5e8
ROM_CLS		equ	$ccce
ROM_LORES0	equ	$d9ed
ROM_KEY		equ	$eb78
ROM_TEXT	equ	$ec21
ROM_HIRES	equ	$ec33
ROM_NMI		equ	$f8b2
ROM_PING	equ	$fa9f
ROM_SHOOT	equ	$fab5
ROM_ZAP		equ	$fae1
ROM_EXPLODE	equ	$facb
ROM_CLICK1	equ	$fb14
ROM_CLICK2	equ	$fb2a

RAM_KEY		equ	$0208
RAM_IRQ		equ	$024a
RAM_STATUS	equ	$026a
RAM_PAPER	equ	$026b
RAM_INK		equ	$026c
RAM_CLOCK	equ	$0276
RAM_HIRES	equ	$a000
RAM_TEXT	equ	$bf68
RAM_LORES	equ	$bb80
RAM_CHARSET	equ	$b400

;--------------------------------------------------------------------

RAM_ZEROPAGE	equ	$40
RAM_START	equ	$1000
RAM_HEAP	equ	$2000

HEAP_TABLEY	equ	RAM_HEAP + $0100

HEAP_SNAKE_S	equ	RAM_HEAP + $0200
HEAP_SNAKE_X	equ	RAM_HEAP + $0300
HEAP_SNAKE_Y	equ	RAM_HEAP + $0400
HEAP_SNAKE_H	equ	RAM_HEAP + $0500
HEAP_SNAKE_V	equ	RAM_HEAP + $0600

KEY_SPACE	equ	132
KEY_ESCAPE	equ	169
KEY_UP		equ	156
KEY_DOWN	equ	180
KEY_LEFT	equ	172
KEY_RIGHT	equ	188
KEY_RETURN	equ	175

SCREEN_WIDTH	equ	40
SCREEN_HEIGHT	equ	28

DEFAULT_PAPER	equ	$16
DEFAULT_INK	equ	$04

;--------------------------------------------------------------------

STATUS_EXIT	equ	0
STATUS_GAMEOVER	equ	1
STATUS_RESET	equ	2
STATUS_LOOP 	equ	3

CODE_GROUND	equ	' '
CODE_FIRE	equ	'|'
CODE_PLAYER	equ	'o'
CODE_SNAKE	equ	96
CODE_MUSHROOM	equ	'a'

PLAYER_X	equ	18
PLAYER_Y	equ	26
PLAYER_LIVES	equ	3
PLAYER_XMIN	equ	2
PLAYER_XMAX	equ	37
PLAYER_YMIN	equ	19
PLAYER_YMAX	equ	26

SNAKE_RIGHT	equ	$00
SNAKE_LEFT	equ	$01
SNAKE_ON	equ	$01
SNAKE_OFF	equ	$00
SNAKE_LENGTH	equ	24
SNAKE_XMIN	equ	2
SNAKE_XMAX	equ	37
SNAKE_YMIN	equ	1
SNAKE_YMAX	equ	27

MUSHROOM_XMIN	equ	3
MUSHROOM_XMAX	equ	32
MUSHROOM_YMIN	equ	2
MUSHROOM_YMAX	equ	24
MUSHROOM_NUMBER	equ	32
MUSHROOM_MAX	equ	4

FRAME_PLAYER	equ	4

PAUSE_SHORT	equ	1
PAUSE_LONG	equ	200

SCORE_SIZE	equ	6

X_LIVES		equ	2
X_SCORE		equ	16
X_HISCORE	equ	33

;--------------------------------------------------------------------

		bss
		org	RAM_ZEROPAGE

;--------------------------------------------------------------------

; global variables
_seed		ds	1
_status		ds	1
_frame		ds	1
_screen		ds	2

_player_x	ds	1
_player_y	ds	1
_player_fire	ds	1
_player_fire_x	ds	1
_player_fire_y	ds	1
_player_lives	ds	1

_snake_number	ds	1
_snake_head	ds	1
_snake_tail	ds	1
_snake_x	ds	1
_snake_y	ds	1
_snake_mode	ds	1

_mushroom	ds	16

;--------------------------------------------------------------------

; void Poke(x,y,code)
; BYTE Peek(x,y)

screen_x	ds	1
screen_y	ds	1
screen_code	ds	1

;--------------------------------------------------------------------

                code
                org     RAM_START - 14

;--------------------------------------------------------------------

Header		db	$16,$16,$16,$24		;
		db	$00			; integers flag (true if != 0x00)
		db	$00			; strings flag  (true if != 0x00)
		db	$80			; type          (0x00 for BASIC, 0x40 for arrays, 0x80 for data)
		db	$01			; auto flag     (true if != 0x00)
  		db	End / 256		; end address   (MSB)
  		db	End % 256		; end address   (LSB)
 		db	Start / 256		; start address (MSB)
  		db	Start % 256		; start address (LSB)
		db	$00			; unused
		db	$00			; filename

;--------------------------------------------------------------------

Start		lda	#$0a
		sta	RAM_STATUS

		lda	#RAM_LORES % 256
		sta	_screen
		lda	#RAM_LORES / 256
		sta	_screen + 1

		ldx	#$00
		stx	RAM_LORES + 35

startup2	lda	_screen
		sta	HEAP_TABLEY,x
		lda	_screen + 1
		sta	HEAP_TABLEY + 1,x

		clc
		lda	_screen
		adc	#40
		sta	_screen
		bcc	startup3
		inc	_screen + 1

startup3	inx
		inx
		cpx	#2 * 28
		bne	startup2

		lda	#$00		
		tax

startup4	sta	_mushroom,x

		inx
		cpx	#$08
		bne	startup4

		tax

startup5	lda	RAM_CHARSET + (94 * 8),x
		sta	_mushroom + 8,x
		sta 	RAM_CHARSET + (CODE_MUSHROOM * 8) + 0,x

		lda	_mushroom + 6,x
		sta 	RAM_CHARSET + (CODE_MUSHROOM * 8) + 8,x

		lda	_mushroom + 4,x
		sta 	RAM_CHARSET + (CODE_MUSHROOM * 8) + 16,x
				
		lda	_mushroom + 2,x
		sta 	RAM_CHARSET + (CODE_MUSHROOM * 8) + 24,x

		inx
		cpx	#$08
		bne	startup5

		lda	#$01
		sta	_frame

		lda	#STATUS_GAMEOVER
		sta	_status

;--------------------------------------------------------------------

; void Irq_Enable()

Irq_Enable	sei

		lda	#$4c			; jmp $HHLL
		sta	RAM_IRQ
		lda	#Irq_Handler % 256
		sta	RAM_IRQ + 1
		lda	#Irq_Handler / 256
		sta	RAM_IRQ + 2			

		cli

main99		clc
		bcc	main99

;--------------------------------------------------------------------

; void Irq_Hanlder()

Irq_Handler	sei

		php
		pha
		txa
		pha
		tya
		pha

Loop		lda	_status
		cmp	#STATUS_GAMEOVER	; _status = GAMEOVER ?
		bne	loop10

		dec	_frame
		bne	loop99

		lda	#PAUSE_SHORT
		sta	_frame

		lda	RAM_KEY
		cmp	#KEY_RETURN		; KEY_RETURN ?
		bne	loop99

		lda	RAM_CLOCK		; randomize
		sta	_seed

		ldx	#SCORE_SIZE
		lda	#'0'
loop1		sta	RAM_LORES + X_SCORE,x
		dex
		bne	loop1

		lda	#PLAYER_X
		sta	_player_x		; _player_x = PLAYER_X

		lda	#PLAYER_Y
		sta	_player_y		; _player_y = PLAYER_Y

		lda	#$00
		sta	_player_fire		; _player_fire = FALSE

		lda	#PLAYER_LIVES
		sta	_player_lives		; _player_lives = PLAYER_LIVES

		lda	#STATUS_RESET
		sta	_status
		bne	loop99

loop10		cmp	#STATUS_RESET		; _status = RESET ?
		bne	loop20

		lda	#PAUSE_SHORT
		sta	_frame

		jsr	Snake_Reset		; Player_Reset()
		jsr	Field			; Field()

		lda	#STATUS_LOOP
		sta	_status
		bne	loop99

loop20		cmp	#STATUS_LOOP		; _status = LOOP ?
		bne	loop99

		dec	_frame
		bne	loop99

		lda	#FRAME_PLAYER
		sta	_frame

		jsr	Player_Move		; Player_Move()
		jsr	Player_Fire		; Player_Fire()

		jsr	Snake_Move		; Snake_Move()

loop99		pla
		tay
		pla
		tax
		pla
		plp

		cli
		rti

;--------------------------------------------------------------------

; void Field()

Field		lda	#DEFAULT_PAPER
		sta	RAM_PAPER
		
		lda	#DEFAULT_INK
		sta	RAM_INK

		jsr	ROM_CLS

		ldx	#MUSHROOM_NUMBER

field1		jsr	Mushroom
		dex
		bne	field1

		rts

;--------------------------------------------------------------------

; void Mushroom()

Mushroom	jsr	Rnd			; x = Rnd()

target2		cmp	#MUSHROOM_XMAX + 1
		bcc	target3			; x < MUSHROOM_XMAX + 1
		sec
		sbc	#MUSHROOM_XMAX		; x -= MUSHROOM_XMAX
		bne	target2
target3		clc
		adc	#MUSHROOM_XMIN		; x += MUSHROOM_XMIN
		sta	screen_x

		jsr	Rnd			; y = Rnd()

target4		cmp	#MUSHROOM_YMAX + 1
		bcc	target5			; y < MUSHROOM_YMAX + 1
		sec
		sbc	#MUSHROOM_YMAX		; y -= MUSHROOM_YMAX
		bne	target4
target5		clc
		adc	#MUSHROOM_YMIN		; y += MUSHROOM_YMIN
		sta	screen_y

		jsr	Peek			; Peek(x,y)
		cmp	#CODE_GROUND
		bne	Mushroom

		lda	#CODE_MUSHROOM
		sta	screen_code
		jsr	Poke			; Poke(x,y,CODE_MUSHROOM)

		rts

;--------------------------------------------------------------------

; void Player_Move()

Player_Move	lda	#$08			; flash off
		sta	RAM_LORES + X_LIVES - 1

		clc
		lda	_player_lives
		adc	#'0'
		sta	RAM_LORES + X_LIVES

		ldx	_player_x
		ldy	_player_y

		lda	#CODE_GROUND
		sta	screen_code
		stx	screen_x
		sty	screen_y
		jsr	Poke			; Poke(player_x,player_y,CODE_GROUND)

		jsr	Key

		stx	screen_x
		sty	screen_y
		jsr	Peek			; Peek(player_x',player_y')

		cmp	#CODE_GROUND
		bne	player_move1

		stx	_player_x		; player_x = player_x'
		sty	_player_y		; player_y = player_y'

player_move1	lda	#CODE_PLAYER
		sta	screen_code
		lda	_player_x
		sta	screen_x
		lda	_player_y
		sta	screen_y
		jmp	Poke			; Poke(player_x,player_y,CODE_PLAYER)

;--------------------------------------------------------------------

; void Player_Fire()

Player_Fire	lda	_player_fire
		beq	player6

		jsr	player10

		cmp	#CODE_FIRE
		bne	player1

		lda	#CODE_GROUND
		sta	screen_code
		jsr	Poke			; Poke(_player_fire_x,_player_fire_y,CODE_GROUND)

player1		dey				; _player_fire_y--
		sty	_player_fire_y

		cpy	#0
		bne	player2

		dec	_player_fire		; _player_fire = FALSE
		rts

player2		jsr	player10

		cmp	#CODE_GROUND
		bne	player3

		lda	#CODE_FIRE
		sta	screen_code
		jmp	Poke			; Poke(_player_fire_x,_player_fire_y,CODE_FIRE)

player3		cmp	#CODE_SNAKE
		bne	player4

		jsr	Snake_Split		; Snake_Split(_player_fire_x,_player_fire_y)

		lda	#CODE_MUSHROOM
		bne	player5

player4		cmp	#CODE_MUSHROOM
		bcc	player6
		cmp	#CODE_MUSHROOM + MUSHROOM_MAX
		bcs	player6

		clc
		adc	#$01
		cmp	#CODE_MUSHROOM + MUSHROOM_MAX
		bne	player5
		lda	#CODE_GROUND
player5		sta	screen_code
		jsr	Poke			; Poke(_player_fire_x,_player_fire_y,CODE_GROUND)

		jsr	Score			; Score()

		dec	_player_fire		; _player_fire = FALSE
player6		rts

player10	ldx	_player_fire_x
		ldy	_player_fire_y
		stx	screen_x
		sty	screen_y
		jmp	Peek			; Peek(_player_fire_x,_player_fire_y)

;--------------------------------------------------------------------

; void Snake_Reset()

Snake_Reset	lda	#SNAKE_XMIN + (SNAKE_XMAX - SNAKE_XMIN) / 2
		sta	_snake_x		; _snake_x = SNAKE_XMIN + (SNAKE_XMAX - SNAKE_XMIN) / 2

		lda	#SNAKE_YMIN
		sta	_snake_y		; _snake_y = SNAKE_YIN

		ldx	#$00

snake_reset1	lda	#SNAKE_ON
		sta	HEAP_SNAKE_S,x

		lda	_snake_x
		sta	HEAP_SNAKE_X,x

		lda	_snake_y
		sta	HEAP_SNAKE_Y,x

		lda	#$00
		sta	HEAP_SNAKE_V,x

		lda	#SNAKE_RIGHT
		sta	HEAP_SNAKE_H,x
	
		inx	
		cpx	#SNAKE_LENGTH
		bne	snake_reset1
	
		lda	#SNAKE_OFF
		sta	HEAP_SNAKE_S,x

		rts

;--------------------------------------------------------------------

; void Snake_Move()

Snake_Move	ldx	#$00			; i = 0
		stx	_snake_number		; _snake_number = 0

snake_move1	lda	HEAP_SNAKE_S,x		; HEAP_SNAKE_S[i] == SNAKE_ON ?
		cmp	#SNAKE_ON
		beq	snake_move2

		inx
		cpx	#SNAKE_LENGTH
		beq	snake_move5
		bne	snake_move1

snake_move2	stx	_snake_head		; _snake_head = i

snake_move3	inx
		lda	HEAP_SNAKE_S,x		; HEAP_SNAKE_S[i] == SNAKE_OFF ?
		cmp	#SNAKE_OFF
		bne	snake_move3		; HEAP_SNAKE_S[SNAKE_LENGTH] = SNAKE_OFF

		stx	_snake_tail

		inc	_snake_number		; _snake_number++

		ldx	_snake_tail
		jsr	snake_crawl97
		jsr	Peek			; Peek(x[tail],y[tail])

		cmp	#CODE_SNAKE
		bne	snake_move4

		lda	#CODE_GROUND
		jsr	snake_crawl99		; Poke(HEAP_SNAKE_X[tail],HEAP_SNAKE_Y[tail],CODE_GROUND)

snake_move4	dex

		lda	#CODE_SNAKE
		jsr	snake_crawl99		; Poke(HEAP_SNAKE_X[i],HEAP_SNAKE_Y[i],CODE_SNAKE)

		cpx	_snake_head
		bne	snake_move4

		jsr	Snake_Crawl

		ldx	_snake_tail
		stx	_snake_head		; _snake_head = _snake_tail

		cpx	#SNAKE_LENGTH
		bne	snake_move1		; _snake_tail == SNAKE_LENGTH ?

snake_move5	lda	_snake_number
		beq	snake_move6		; _snake_number == 0 ?
		rts

snake_move6	jmp	Snake_Reset

;--------------------------------------------------------------------

; void Snake_Split(_player_fire_x,_player_fire_y)

Snake_Split	ldx	#$00

snake_split1	lda	HEAP_SNAKE_X,x
		cmp	_player_fire_x
		bne	snake_split2		; HEAP_SNAKE_X[i] == _player_fire_x ?

		lda	HEAP_SNAKE_Y,x
		cmp	_player_fire_y		; HEAP_SNAKE_Y[i] == _player_fire_y ?
		bne	snake_split2

		lda	#SNAKE_OFF
		sta	HEAP_SNAKE_S,x		; HEAP_SNAKE_S[i] = SNAKE_OFF

		inx

		lda	#CODE_GROUND
		jmp	snake_crawl99		; Poke(HEAP_SNAKE_X[i + 1],HEAP_SNAKE_Y[i + 1],CODE_GROUND)

snake_split2	inx
		cpx	#SNAKE_LENGTH
		bne	snake_split1
		rts

;--------------------------------------------------------------------

; void Snake_Crawl(head,tail)

Snake_Crawl	ldx	_snake_tail

snake_crawl1	lda	HEAP_SNAKE_X - 1,x	; scroll body
		sta	HEAP_SNAKE_X,x

		lda	HEAP_SNAKE_Y - 1,x
		sta	HEAP_SNAKE_Y,x
		
		dex
		cpx	_snake_head
		bne	snake_crawl1

		lda	HEAP_SNAKE_V,x		; HEAP_SNAKE_V[head] == 0 ?
		beq	snake_crawl10

		dec	HEAP_SNAKE_V,x		; HEAP_SNAKE_V[head]--
		bne	snake_crawl10

		lda	HEAP_SNAKE_Y,x		; HEAP_SNAKE_Y[head] == bottom ?
		cmp	#SNAKE_YMAX
		bne	snake_crawl2

		lda	#SNAKE_YMIN
		sta	HEAP_SNAKE_Y,x		; HEAP_SNAKE_S[head] = SNAKE_OFF

		jsr	snake_crawl97
		lda	#CODE_GROUND
		sta	screen_code
		jmp	Poke			; Poke(HEAP_SNAKE_X[tail],HEAP_SNAKE_Y[tail],CODE_GROUND)

snake_crawl2	jsr	snake_crawl97
		inc	screen_y
		jsr	Peek			; Peek(x[head],y[head] + 1)

		cmp	#CODE_PLAYER
		beq	snake_crawl50

		cmp	#CODE_GROUND
		bne	snake_crawl3
	
		jmp	snake_crawl98

snake_crawl3	inc	HEAP_SNAKE_V,x		; HEAP_SNAKE_V[head]++

snake_crawl10	lda	HEAP_SNAKE_H,x
		cmp	#SNAKE_RIGHT
		bne	snake_crawl20		

		lda	HEAP_SNAKE_X,x		; HEAP_SNAKE_H[head] == RIGHT
		cmp	#SNAKE_XMAX
		beq	snake_crawl30		; HEAP_SNAKE_X[head] == SNAKE_XMAX ?

		jsr	snake_crawl97
		inc	screen_x
		jsr	Peek			; Peek(x[head] + 1,y[head])

		cmp	#CODE_PLAYER
		beq	snake_crawl50
	
		cmp	#CODE_GROUND
		bne	snake_crawl30

		jmp	snake_crawl98

snake_crawl20	lda	HEAP_SNAKE_X,x		; HEAP_SNAKE_H[head] == LEFT
		cmp	#SNAKE_XMIN
		beq	snake_crawl30		; HEAP_SNAKE_X[head] == SNAKE_XMIN ?

		jsr	snake_crawl97
		dec	screen_x
		jsr	Peek			; Peek(x[head] - 1,y[head])

		cmp	#CODE_PLAYER
		beq	snake_crawl50

		cmp	#CODE_GROUND
		bne	snake_crawl30

		jmp	snake_crawl98

snake_crawl30	inc	HEAP_SNAKE_V,x		; HEAP_SNAKE_V[head]++

		lda	HEAP_SNAKE_H,x
		eor	#$01
		sta	HEAP_SNAKE_H,x		; HEAP_SNAKE_H[head] = ~HEAP_SNAKE_H[head]
		rts

snake_crawl50	lda	#CODE_GROUND
		sta	screen_code
		jsr	Poke			; Poke(_player_x,_player_y,CODE_GROUND)

		lda	#$0c			; flash on
		sta	RAM_LORES + X_LIVES - 1

		lda	#PLAYER_XMIN + (PLAYER_XMAX - PLAYER_XMIN) / 2
		sta	_player_x

		lda	#PLAYER_YMAX
		sta	_player_y

		dec	_player_lives		; _player_lives--
		beq	snake_crawl51

		lda	#PAUSE_LONG
		sta	_frame
		rts

snake_crawl51	jsr	HiScore			; HiScore()

		lda	#STATUS_GAMEOVER
		sta	_status
		rts

snake_crawl97	lda	HEAP_SNAKE_X,x
		sta	screen_x
		lda	HEAP_SNAKE_Y,x
		sta	screen_y
		rts

snake_crawl98	lda	screen_x
		sta	HEAP_SNAKE_X,x
		lda	screen_y
		sta	HEAP_SNAKE_Y,x
		rts

snake_crawl99	sta	screen_code
		lda	HEAP_SNAKE_X,x
		sta	screen_x
		lda	HEAP_SNAKE_Y,x
		sta	screen_y
		jmp	Poke			; Poke(x[i],y[i],CODE_SNAKE)

;--------------------------------------------------------------------

; void Key()

Key		lda	RAM_KEY

Key_right	cmp	#KEY_RIGHT
		bne	Key_left

		cpx	#PLAYER_XMAX
		beq	Key_other

		inx
		rts

Key_left	cmp	#KEY_LEFT
		bne	Key_up

		cpx	#PLAYER_XMIN
		beq	Key_other

		dex
		rts

Key_up		cmp	#KEY_UP
		bne	Key_down

		cpy	#PLAYER_YMIN
		beq	Key_other

		dey
		rts

Key_down	cmp	#KEY_DOWN
		bne	Key_space

		cpy	#PLAYER_YMAX
		beq	Key_other

		iny
		rts

Key_space	cmp	#KEY_SPACE
		bne	Key_other

		lda	_player_fire
		bne	Key_other		; _player_fire == TRUE ?

		stx	_player_fire_x		; _player_fire_x = _player_x
		sty	_player_fire_y		; _player_fire_y = _player_y

		inc	_player_fire		; _player_fire = TRUE

Key_other	rts

;--------------------------------------------------------------------

; void Score()

Score		sec
		ldx	#SCORE_SIZE - 1

score1		lda	RAM_LORES + X_SCORE,x
		adc	#$00
		cmp	#'0' + 10
		bne	score2
		lda	#'0'
		sec
score2		sta	RAM_LORES + X_SCORE,x
		dex
		bne	score1

		rts

;--------------------------------------------------------------------

; void HiScore()

HiScore		ldy	#$00

hiscore1	lda	RAM_LORES + X_HISCORE,y
		cmp	RAM_LORES + X_SCORE,y
		bcc	hiscore2		; X_HISCORE[i] < X_SCORE[i] ?

		iny
		cpy	#SCORE_SIZE
		bne	hiscore1
		rts

hiscore2	ldy	#SCORE_SIZE

hiscore3	lda	RAM_LORES + X_SCORE,y
		sta	RAM_LORES + X_HISCORE,y	; X_HISCORE[i] = X_SCORE[i]
		dey
		bne	hiscore3
		rts

;--------------------------------------------------------------------


; void Poke(x,y,code)

Poke		tya
		pha

		jsr	GetPosition

		lda	screen_code
		sta	(_screen),y
		
		pla
		tay
		rts

;--------------------------------------------------------------------

; BYTE Peek(x,y)

Peek		tya
		pha

		jsr	GetPosition

		lda	(_screen),y
		sta	screen_code
		
		pla
		tay

		lda	screen_code
		rts

;--------------------------------------------------------------------

GetPosition	lda	screen_y
		asl	a
		tay	
		lda	HEAP_TABLEY,y
		sta	_screen
		lda	HEAP_TABLEY + 1,y
		sta	_screen + 1
		ldy	screen_x
		rts

;--------------------------------------------------------------------

; char Rnd()

Rnd		txa
		pha
		tya
		pha

		lda	_seed
		and	#$b8
		ldx	#$05
		ldy	#$00

rnd1		asl	a
		bcc	rnd2
		iny
rnd2		dex
		bne	rnd1
		tya
		lsr	a
		lda	_seed
		rol	a
		sta	_seed

		pla
		tay
		pla
		tax

		lda	_seed
		rts

;--------------------------------------------------------------------

End
